#!/usr/bin/env python

from optparse import OptionParser, OptionGroup, SUPPRESS_HELP
import os
import sys

from mapdamage.version import __version__
from mapdamage.rscript import check_R_lib

def file_exist(filename):
    if os.path.exists(filename) and not os.path.isdir(filename):
        return True
    elif filename == "-":
        return True
    else:
        sys.stderr.write("Error: '%s' is not a valid file\n" % (filename))
        return None


def whereis(program):
    for path in os.environ.get('PATH', '').split(':'):
        if os.path.exists(os.path.join(path, program)) and \
            not os.path.isdir(os.path.join(path, program)):
            return os.path.join(path, program)
    return None



def check_py_version():
    req_version = (2, 6)
    cur_version = sys.version_info

    if cur_version >= req_version:
        return True
    else:
        sys.stderr.write("Your Python interpreter is too old."\
            "Please consider upgrading to at least %d.%d\n" % (req_version[0], req_version[1]))
        return None


def options():  
    parser = OptionParser("%prog [options] -i BAMfile -r reference.fasta\n\nUse option -h or --help for help", version=__version__, \
            epilog="report bugs to aginolhac@snm.ku.dk, MSchubert@snm.ku.dk or jonsson.hakon@gmail.com")

    args = OptionGroup(parser, "Input files")
    args.add_option("-i", "--input", help="SAM/BAM file, must contain a valid header, use '-' for reading a BAM from stdin", \
          action="store", type="string", dest="filename")
    args.add_option("-r", "--reference", help="Reference file in FASTA format", \
          action="store", dest="ref")

    parser.add_option_group(args)
    group = OptionGroup(parser, "General options")
    group.add_option("-n", "--downsample", help = "Downsample to a randomly selected fraction of the reads (if 0 < DOWNSAMPLE < 1), or " \
                     "a fixed number of randomly selected reads (if DOWNSAMPLE >= 1). By default, no downsampling is performed.",
                     type = float, default = None)
    group.add_option("--downsample-seed", help = "Seed value to use for downsampling. See documentation for py module 'random' for default behavior.",
                     type = int, default = None)
    group.add_option("--merge-reference-sequences", help = "Ignore referece sequence names when tabulating reads (using '*' instead). "
                     "Useful for alignments with a large number of reference sequnces, which may otherwise result in excessive "
                     "memory or disk usage due to the number of tables generated.",
                     default = False, action = "store_true")
    group.add_option("-l", "--length", dest="length", help="read length, in nucleotides to consider [%default]", \
            type = int, default=70,action="store")
    group.add_option("-a", "--around", dest="around", help="nucleotides to retrieve before/after reads [%default]", \
            type = int, default=10,action="store")
    group.add_option("-Q", "--min-basequal", dest="minqual", help="minimun base quality Phred score considered, Phred-33 assumed [%default]", \
            type = int, default=0, action="store")
    group.add_option("-d", "--folder", help="folder name to store results [results_FILENAME]", \
          action="store", type="string", dest="folder")
    group.add_option("-f", "--fasta", dest="fasta", help="Write alignments in a FASTA file", \
          default=False,action="store_true")
    group.add_option("--plot-only", dest="plot_only", help="Run only plotting from a valid result folder", \
          default=False,action="store_true")
    group.add_option("-q", "--quiet", dest="quiet", help="Disable any output to stdout", \
          default=False,action="store_true")
    group.add_option("-v", "--verbose", dest="verbose", help="Display progression information during parsing", \
          default=False,action="store_true")
    group.add_option("--mapdamage-modules", dest="mapdamage_modules", help="Override the system wide installed mapDamage module", \
          default=None)
    group.add_option("--no-plot", dest="no_r", help=SUPPRESS_HELP, default=False, action="store_true")
    parser.add_option_group(group)

    # options for plotting damage patterns
    group2 = OptionGroup(parser, "Options for graphics")
    group2.add_option("-y", "--ymax", dest="ymax", \
           help="graphical y-axis limit for nucleotide misincorporation frequencies [%default]", type = float, \
           default=0.3,action="store")
    group2.add_option("-m", "--readplot", dest="readplot", \
           help="read length, in nucleotides, considered for plotting nucleotide misincorporations [%default]", \
           type = int, default=25, action="store")
    group2.add_option("-b", "--refplot", dest="refplot", \
          help="the number of reference nucleotides to consider for ploting base composition in the region located upstream "
          "and downstream of every read [%default]", type= int, default=10, action="store")
    group2.add_option("-t", "--title", dest="title", \
          help="title used for plots [%default]", \
          type="string", default="",action="store")
    parser.add_option_group(group2)

    # Then the plethora of optional options for the statistical estimation ..
    group3 = OptionGroup(parser,"Options for the statistical estimation")
    group3.add_option("", "--rand", dest="rand", \
            help="Number of random starting points for the likelihood optimization  [%default]", type = int, default=30, action="store")
    group3.add_option("", "--burn", dest="burn", \
            help="Number of burnin iterations  [%default]", type = int, default=10000,action="store")
    group3.add_option("", "--adjust", dest="adjust", \
            help="Number of adjust proposal variance parameters iterations  [%default]", type = int, default=10, action="store")
    group3.add_option("", "--iter", dest="iter", \
            help="Number of final MCMC iterations  [%default]", type = int, default=50000, action="store")
    group3.add_option("", "--forward", dest="forward", \
            help="Using only the 5' end of the seqs  [%default]", default=False, action="store_true")
    group3.add_option("", "--reverse", dest="reverse", \
            help="Using only the 3' end of the seqs  [%default]", default=False, action="store_true")
    group3.add_option("", "--var-disp", dest="var_disp", \
            help="Variable dispersion in the overhangs  [%default]", default=False,action="store_true")
    group3.add_option("", "--jukes-cantor", dest="jukes_cantor", \
            help="Use Jukes Cantor instead of HKY85  [%default]", default=False,action="store_true")
    group3.add_option("", "--diff-hangs", dest="diff_hangs", \
            help="The overhangs are different for 5' and 3'  [%default]", default=False, action="store_true")
    group3.add_option("", "--fix-nicks" , dest="fix_nicks", \
            help="Fix the nick frequency vector (Only C.T from the 5' end and G.A from the 3' end)  [%default]", default=False, action="store_true")
    group3.add_option("", "--use-raw-nick-freq" , dest="use_raw_nick_freq", \
            help="Use the raw nick frequency vector without smoothing  [%default]", default=False, action="store_true")
    group3.add_option("", "--single-stranded", dest="single_stranded", \
            help="Single stranded protocol [%default]", default=False, action="store_true")
    group3.add_option("", "--theme-bw", dest="theme_bw", \
            help="Use black and white theme in post. pred. plot [%default]", default=False, action="store_true")
    group3.add_option("", "--seq-length", dest="seq_length", \
            help="How long sequence to use from each side [%default]", type = int, default=12, action="store")
    group3.add_option("--stats-only", dest="stats_only", help="Run only statistical estimation from a valid result folder", \
          default=False, action="store_true")
    group3.add_option("--no-stats", help="Disabled statistical estimation, active by default", default=False, action="store_true")
    group3.add_option("--check-R-packages", help="Check if the R modules are working", default=False, action="store_true")
    parser.add_option_group(group3)

    group4 = OptionGroup(parser,"Options for rescaling of BAM files")
    group4.add_option("--rescale", dest="rescale", help="Rescale the quality scores in the BAM file using the output from the statistical estimation",
          default=False, action="store_true")
    group4.add_option("--rescale-only", dest="rescale_only", help="Run only rescaling from a valid result folder",
          default=False, action="store_true")
    group4.add_option("--rescale-out", dest="rescale_out", help="Write the rescaled BAM to this file",
          default=None, action="store")
    group4.add_option("--rescale-length-5p", dest="rescale_length_5p",
                      help="How many bases to rescale at the 5' termini; defaults to --seq-length.", type=int, action="store")
    group4.add_option("--rescale-length-3p", dest="rescale_length_3p",
                      help="How many bases to rescale at the 5' termini; defaults to --seq-length.", type=int, action="store")
    parser.add_option_group(group4)


    #Parse the arguments
    (options, args) = parser.parse_args()

    # check python version
    if not check_py_version():
        return None

    # if the user wants to check the R packages then do that before the option parsing 
    if options.check_R_packages:
        if check_R_lib():
            sys.exit(1)
        else:
            print("All R packages are present")
            sys.exit(0)

    # check general arguments
    if not (options.plot_only or options.stats_only) and not options.filename:
        parser.error('SAM/BAM file not given (-i)')
    if not (options.plot_only or options.ref):
        parser.error('Reference file not given (-r)')
    if not options.plot_only and not options.stats_only:
        if not file_exist(options.filename) or not file_exist(options.ref):
            return None
    if options.downsample is not None:
        if options.downsample <= 0:
            parser.error("-n/--downsample must be a positive value")
        elif options.downsample >= 1:
            options.downsample = int(options.downsample)

    if options.plot_only and not options.folder:
        parser.error('Folder not provided, required with --plot-only')
    if options.stats_only and not options.folder:
        parser.error('Folder not provided, required with --stats-only')
    if options.rescale_only and not options.folder:
        parser.error('Folder not provided, required with --rescale-only')
    if options.rescale_only and not options.filename:
        parser.error('Input bam not provided, required with --rescale-only')
    if options.rescale_only and not options.ref:
        parser.error('Reference not provided, required with --rescale-only')

    if options.verbose and options.quiet:
        parser.error('Cannot use verbose and quiet option at the same time')

    # check options
    if options.length < 0:
        parser.error('length (-l) must be a positive integrer')
    if options.around < 0:
        parser.error('around (-a) must be a positive integrer')
    if options.ymax <= 0 or options.ymax > 1:
        parser.error('ymax (-b) must be an real number beetween 0 and 1')
    if options.readplot < 0:
        parser.error('readplot (-m) must be a positive integrer')
    if options.refplot < 0:
        parser.error('refplot (-b) must be a positive integrer')
    if options.refplot > options.around and not options.plot_only:
        parser.error('refplot (-b) must be inferior to around (-a)')
    if options.readplot > options.length:
        parser.error('readplot (-m) must be inferior to length (-l)')
    if options.minqual < 0 or  options.minqual > 41:
        parser.error('minimal base quality, Phred score, must be within this range: 0 - 41')

    # check statistic options
    if options.forward and options.reverse:
        parser.error('Cannot use only forward end and only reverse end for the statistics')

    # use filename as default for plot titles if not set
    if options.title == "" and options.filename:
        options.title = os.path.splitext(os.path.basename(options.filename))[0]
    # for --plot-only, use the folder name, without results_ as title
    if options.title == "" and not options.filename and options.folder:
        options.title = os.path.splitext(os.path.basename(options.folder))[0].replace("results_", "")

    # check folder
    if not options.folder and options.filename:
        options.folder = "results_"+os.path.splitext(os.path.basename(options.filename))[0]

    # check destination for rescaled bam
    if not options.rescale_out and (options.rescale or options.rescale_only):
        # if there are mulitiple bam files to rescale then pick first one as 
        # the name of the rescaled file
        if isinstance(options.filename,list):
            basename = os.path.basename(options.filename[0])
        else:
            basename = os.path.basename(options.filename)
        with_ext = os.path.splitext(basename)[0] + ".rescaled.bam"
        options.rescale_out = os.path.join(options.folder, with_ext)

    if os.path.isdir(options.folder):
        if not options.quiet and not options.plot_only:
            print("Warning, %s already exists" % options.folder)
        if options.plot_only:
            if not file_exist(options.folder+"/dnacomp.txt") or not file_exist(options.folder+"/misincorporation.txt"):
                parser.error('folder %s is not a valid result folder' % options.folder)
    else:
        os.makedirs(options.folder, mode = 0750)
        if options.plot_only or options.stats_only or options.rescale_only:
            sys.stderr.write("Error, %s does not exist while plot/stats/rescale only was used\n" % options.folder)
            return None

    if options.rescale_length_3p is None:
        options.rescale_length_3p = options.seq_length
    elif not (0 <= options.rescale_length_3p <= options.seq_length):
        parser.error("--rescale-length-3p must be less than or equal to "
                     "--seq-length and greater than zero")

    if options.rescale_length_5p is None:
        options.rescale_length_5p = options.seq_length
    elif not (0 <= options.rescale_length_5p <= options.seq_length):
        parser.error("--rescale-length-5p must be less than or equal to "
                     "--seq-length and greater than zero")

    # check if the Rscript executable is present on the system
    if not whereis('Rscript'):
        print("Warning, Rscript is not in your PATH, plotting is disabled")
        options.no_r = True

    # check the nick frequencies options
    if (options.use_raw_nick_freq + options.fix_nicks + options.single_stranded)>1:
        parser.error("The options --use-raw-nick-freq, --fix-nicks and --single-stranded are mutually exclusive.")

    if check_R_lib():
        # check for R libraries
        print("The Bayesian estimation has been disabled\n")
        options.no_stats = True
        if options.stats_only:
            sys.exit("Cannot use --stats-only with missing R libraries")
        if options.rescale:
            sys.exit("Cannot use --rescale with missing R libraries")
        if options.rescale_only:
            sys.exit("Cannot use --rescale-only with missing R libraries")


    return options




import unittest
import optparse
import rescale
import pysam
import filecmp

def mock_options(filename,rescale_out,folder):
    """Make the options object with nice values for testing"""
    return optparse.Values({
        "filename":filename,
        "rescale_out":rescale_out,
        "verbose":True,
        "folder":folder,
        "quiet":True
        })


class testPairedFile(unittest.TestCase):
    """Tests if rescaling of a paired end file"""
    def test_paired_end_file(self):
        """Test, paired end SAM file"""
        # here is the example
        # >ref
        # 5p CGA AAA CGA 3p
        # 3p GCT TTT GCT 5p
        # >r001/1
        # CGA
        # >r001/2
        # GCA
        # The sam file looks like this
        #@HD VN:1.5 SO:coordinate
        #@SQ SN:ref LN:9
        #r001 163 ref 1 30 3M = 7 9 CGA III MR:f:0 #the normal ones
        #r001 83 ref 7 30 3M = 1 -9 TCG III MR:f:0
        #r002 163 ref 1 30 3M = 7 9 TGA III MR:f:0.9 #With one dam subs
        #r002 83 ref 7 30 3M = 1 -9 CAA III MR:f:0.009
        #r003 83 ref 1 30 3M = 7 9 TGA III #The reverse complement, should not rescale (thus no flags)
        #r003 163 ref 7 30 3M = 1 -9 CAA III
        # 
        #hand calculated the correct rescaled sam file in pe_rescaled_correct.sam
        options = mock_options("rescale_test/pe_test/pe.sam","rescale_test/pe_test/pe_rescaled.sam","rescale_test/pe_test/pe_output/")
        ref = pysam.Fastafile("rescale_test/pe_test/ref.fa")
        rescale.rescale_qual(ref,options,debug=True)
        self.assertTrue(filecmp.cmp("rescale_test/pe_test/pe_rescaled.sam","rescale_test/pe_test/pe_rescaled_correct.sam"))

class testCases(unittest.TestCase):
    """Various cases that failed"""
    def test_longalignshortread(self):
        """Check if fails on an aligment longer than the read"""
        prefix="rescale_test/long_align/"
        options = mock_options(prefix+"pe.sam",prefix+"pe_out.sam",prefix+"pe_output/")
        ref = pysam.Fastafile("rescale_test/pe_test/ref.fa")
        rescale.rescale_qual(ref,options,debug=True)
        #Should run without an error

if  __name__=='__main__':
    unittest.main()



#!/usr/bin/env python

import os
import collections

import mapdamage
from mapdamage.version import __version__


def initialize_mut(ref, length):  
  tab = {}
  for contig in ref:
    tab_contig = tab[contig] = {}
    for end in ('5p','3p'):
      tab_end = tab_contig[end] = {}
      for std in ('+','-'):
        tab_std = tab_end[std] = {}
        for mut in mapdamage.seq.HEADER:
          tab_std[mut] = dict.fromkeys(xrange(length), 0)
  
  return tab


def print_mut(mut, opt, out):
  _print_freq_table(mut, mapdamage.seq.HEADER, opt, out, offset = 1)


def initialize_comp(ref, around, length):
  keys = {"3p" : range(-length, 0) + range(1, around + 1),
          "5p" : range(-around, 0) + range(1, length + 1)}

  tab = {}
  for contig in ref:
    tab_contig = tab[contig] = {}
    for end in ('5p','3p'):
      tab_end = tab_contig[end] = {}
      for std in ('+','-'):
        tab_std = tab_end[std] = {}
        for letters in mapdamage.seq.LETTERS:
          tab_std[letters] = dict.fromkeys(keys[end], 0)

  return tab


def print_comp(comp, opt, out):
  columns = mapdamage.seq.LETTERS + ("Total",)
  _print_freq_table(comp, columns, opt, out)


def initialize_lg():
  tab = {}
  for std in ('+','-'):
    tab[std] = collections.defaultdict(int)

  return tab 


def print_lg(tab, opt, out):
  out.write("# table produced by mapDamage version %s\n" % __version__)
  out.write("# using mapped file %s and %s as reference file\n" % (opt.filename, opt.ref))
  if opt.minqual != 0:
    out.write("# Quality filtering of bases with a Phred score < %d\n" % opt.minqual)
  out.write("# Std: strand of reads\n")
  out.write("Std\tLength\tOccurences \n")
  for std in tab:
    for i in tab[std]:
      out.write("%s\t%d\t%d\n" % (std, i, tab[std][i]))


def check_table_and_warn_if_dmg_freq_is_low(folder):
  """ Returns true if the damage frequencies are too low to allow
  Bayesian estimation of DNA damages, i.e < 1% at first position.
  """
  total = 0.0
  for filename in ("5pCtoT_freq.txt", "3pGtoA_freq.txt"):
    if not os.path.exists(folder+"/"+filename):
      print("Error: Required table has not been created ('%s'), bayesian computation cannot be performed" \
            % filename)
      return True

    with open(os.path.join(folder, filename)) as handle:
      for line in handle:
        freq = line.strip().split('\t')
        if freq[0] == "1":
          total += float(freq[1])
          break
      else:
        print("Error: Could not find pos = 1 in table '%s', bayesian computation cannot be performed" \
              % filename)
        return True

  if total < 0.01:
    print("Warning: DNA damage levels are too low, the Bayesian computation should not be performed (%f < 0.01)\n" % total)

  return False


def _print_freq_table(table, columns, opt, out, offset = 0):
  out.write("# table produced by mapDamage version %s\n" % __version__)
  out.write("# using mapped file %s and %s as reference file\n" % (opt.filename, opt.ref))
  if opt.minqual != 0:
    out.write("# Bases with a Phred score < %d were filtered out\n" % opt.minqual)
  out.write("# Chr: reference from sam/bam header, End: from which termini of DNA sequences, Std: strand of reads\n")
  out.write("Chr\tEnd\tStd\tPos\t%s\n" % ("\t".join(columns)))

  for (reference, ends) in sorted(table.iteritems()):
    for (end, strands) in sorted(ends.iteritems()):
      for (strand, subtable) in sorted(strands.iteritems()):
        subtable["Total"] = {}
        for index in sorted(subtable[columns[0]]):
          subtable["Total"][index] = sum(subtable[letter][index] for letter in mapdamage.seq.LETTERS)

          out.write("%s\t%s\t%s\t%d" % (reference, end, strand, index + offset))
          for base in columns:
            out.write("\t%d" % subtable[base][index])
          out.write("\n")



#!/usr/bin/env python

import mapdamage.parseoptions
import mapdamage.seq
import mapdamage.align
import mapdamage.tables
import mapdamage.composition
import mapdamage.rscript
import mapdamage.rescale
import mapdamage.version



#!/usr/bin/env python

import mapdamage
import itertools
import csv
import subprocess
import sys

def count_ref_comp(read, chrom, before, after, comp):
  """ record basae composition in external genomic regions """
  std = '-' if read.is_reverse else '+'

  _update_table(comp[chrom]['5p'][std], before, xrange(-len(before), 0))
  _update_table(comp[chrom]['3p'][std], after,  xrange(1, len(after) + 1))


def count_read_comp(read, chrom, length, comp):
  """ record base composition of read, discard marked nucleotides """
  std, seq = '+', read.query
  if read.is_reverse:
    std, seq = '-', mapdamage.seq.revcomp(seq)

  _update_table(comp[chrom]['5p'][std], seq,           xrange(1, length + 1))
  _update_table(comp[chrom]['3p'][std], reversed(seq), xrange(-1, - length - 1, -1))


def _update_table(table, sequence, indices):
  for (index, nt) in itertools.izip(indices, sequence):
    if nt in table:
      table[nt][index] += 1


def get_base_comp(filename,destination=False):
    """
    Gets the basecomposition of all the sequences in filename
    and returns the value to destination if given.
    """
    path_to_seqtk = mapdamage.rscript.construct_path("seqtk",folder="seqtk") 
    bases = {"A":0,"C":0,"G":0,"T":0}
    alp = ["A","C","G","T"]
    try:
        proc = subprocess.Popen([path_to_seqtk,"comp",filename],stdout=subprocess.PIPE)
        out = proc.communicate()[0]
        for li in out.splitlines():
            tabs = li.split() # 1 is the ref, 2 is the total and then the base counts A, C, G and T.
            bases["A"] = bases["A"] + int(tabs[2])
            bases["C"] = bases["C"] + int(tabs[3])
            bases["G"] = bases["G"] + int(tabs[4])
            bases["T"] = bases["T"] + int(tabs[5])
    except (OSError, ValueError), error:
        sys.stderr.write("Error: Seqtk failed: %s\n" % (error,))
        sys.exit(1)
    # get the base frequencies
    ba_su = sum(bases.values())
    for ba in alp:
        bases[ba] = float(bases[ba])/float(ba_su)
    if (destination==False):
        return bases
    else:
        # write the results
        fo = open(destination,"w")
        vals = [str(bases[i]) for i in alp]
        fo.write(",".join(alp)+"\n")
        fo.write(",".join(vals)+"\n")
        fo.close()

def read_base_comp(filename):
    """
    Read the base compition from a file created by get_base_comp
    """
    fh = open(filename)
    first_line = True
    for li in fh:
        li = li.rstrip()
        lp = li.split()
        if first_line:
            header = lp
            first_line = False 
        else:
            body = lp
    bases = {}
    for ba,per in zip(header,body):
        bases[ba] = per
    return bases



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from distutils.core import setup
from distutils.command.install import install as DistutilsInstall
import os
import subprocess

def compile_seqtk():
    """Compiling the seqtk toolkit"""
    old_wd = os.getcwd()
    new_wd = os.path.join(old_wd,"mapdamage","seqtk")
    os.chdir(new_wd)
    if not os.path.isfile("seqtk.c"):
        raise SystemExit("Cannot find seqtk.c")
    if (os.path.isfile("seqtk")):
        os.system("rm seqtk")
    xs = os.system("make -f Makefile")
    os.chdir(old_wd)
    if (xs != 0):
        raise SystemExit("Cannot compile seqtk")


def setup_version():
    if not os.path.exists(".git"):
        # Release version, no .git folder
        return

    try:
        version = subprocess.check_output(("git", "describe", "--always", "--tags", "--dirty"))
        with open(os.path.join("mapdamage", "_version.py"), "w") as handle:
            handle.write("#!/usr/bin/env python\n")
            handle.write("__version__ = %r\n" % (version.strip(),))
    except (subprocess.CalledProcessError, OSError), error:
        raise SystemExit("Could not determine mapDamage version: %s" % (error,))


class compileInstall(DistutilsInstall):
    # extension of the class to account for an extra compiling step
    def run(self):
        self.record=""
        setup_version()
        compile_seqtk()
        DistutilsInstall.run(self)
        # fixing the permission problem of seqtk
        files = self.get_outputs()
        for fi in files:
            if fi.endswith("seqtk/seqtk"):
                os.chmod(fi,0755)





setup(
    cmdclass={'install': compileInstall},
    name='mapdamage',
    version='2.0.8',
    author='Aurélien Ginolhac, Mikkel Schubert, Hákon Jónsson',
    author_email='MSchubert@snm.ku.dk, jonsson.hakon@gmail.com',
    packages=['mapdamage'],
    package_data={'mapdamage': ['Rscripts/*.R','Rscripts/stats/*.R','seqtk/seqtk']},
    scripts=['bin/mapDamage'],
    url='https://github.com/ginolhac/mapDamage',
    license='LICENSE.txt',
    description='mapDamage tracks and quantify DNA damage pattern among ancient DNA sequencing reads generated by Next-Generation Sequencing platforms',
    long_description=open('README.md').read()
)



#!/usr/bin/env python
import sys
import string

# from Martin Kircher, to complement DNA
TABLE = string.maketrans('TGCAMRWSYKVHDBtgcamrwsykvhdb', \
    'ACGTKYWSRMBDHVacgtkywsrmbdhv')

LETTERS = ("A", "C", "G", "T")
MUTATIONS = ('G>A', 'C>T', 'A>G', 'T>C', 'A>C', 'A>T', 'C>G', 'C>A', 'T>G',
             'T>A', 'G>C', 'G>T', 'A>-', 'T>-', 'C>-', 'G>-', '->A', '->T', 
             '->C', '->G', 'S')
HEADER = LETTERS + ("Total", ) + MUTATIONS


def write_fasta(read, ref, seq, refseq, start, end, before, after, fout):
  std = '-' if read.is_reverse else '+'

  # output coordinate in 1-based offset for 5'
  if len(before) > 0:
    fout.write(">%s\n%s\n" % (ref, before))
  fout.write(">%s:%d-%d\n%s\n>%s_%s\n%s\n" % (ref, start+1, end, refseq, read.qname, std, seq))
  if len(after) > 0:
    fout.write(">%s\n%s\n" % (ref, after))


def revcomp(seq):
  """ return reverse complemented string """
  return seq.translate(TABLE)[::-1]


def record_lg(read, coordinate, tab):
  """ record global length distribution
  don't record paired reads as they are normally not used for aDNA """
  std = '-' if read.is_reverse else '+'
  
  length = (max(coordinate) - min(coordinate))
  if not read.is_paired:
    tab[std][length] = tab[std][length] + 1
    
  return tab


def read_fasta_index(filename):
  """ from a fasta index file, fai, return dictionary of references:lengths """
  def print_err(msg, filename, line):
    sys.stderr.write("Error: %s\n" % msg)
    sys.stderr.write("       Filename: %s\n" % filename)
    sys.stderr.write("       Line:     %s\n" % repr(line))
  
  fai = {}
  with open(filename, 'r') as handle:
    for line in handle:
      ref = line.split("\t")
      if len(ref) != 5:
        print_err("Line in fasta index contains wrong number of fields, found %i, expected 5:" \
            % len(ref), filename, line)
        return None

      try:
        fai[ref[0]] = int(ref[1])
      except ValueError:
        print_err("Column 2 in FASTA index did not contain a number, found '%s':" % ref[1], filename, line)
        return None

  return fai


def compare_sequence_dicts(fasta_dict, bam_dict):
  """Compares a FASTA and BAM sequence dictionary, and prints any differences.
  Returns true if all required sequences are found, false otherwise."""
  if fasta_dict == bam_dict:
    return True

  sys.stderr.write("Sequence dictionaries in FASTA/BAM files differ:\n")
  common = set(fasta_dict) & set(bam_dict)
  if not common:
    sys.stderr.write("FATAL ERROR: No sequences in common!\n")
    return False

  # Check that the lengths of required sequences match (fatal error)
  different = []
  for key in sorted(common):
    if fasta_dict[key] != bam_dict[key]:
      different.append((key, fasta_dict[key], bam_dict[key]))

  if different:
    sys.stderr.write("FATAL ERROR: Length of required sequences differ:\n")
    for values in different:
      sys.stderr.write("    - %s: %i bp vs %i bp\n" % values)

  # Check for sequences only found in the BAM file (fatal errors)
  bam_only = set(bam_dict) - common
  if bam_only:
    sys.stderr.write("FATAL ERROR: Sequences missing from FASTA dictionary:\n")
    for key in bam_only:
      sys.stderr.write("    - %s = %i bp\n" % (key, bam_dict[key]))

  # Check for sequences only found in the BAM file (fatal errors)
  fasta_only = set(fasta_dict) - common
  if fasta_only:
    sys.stderr.write("WARNING: FASTA file contains extra sequences:\n")
    for key in fasta_only:
      sys.stderr.write("    - %s = %i bp\n" % (key, fasta_dict[key]))

  sys.stderr.write("\n")

  return not (different or bam_only)



#!/usr/bin/env python

import os
import subprocess
from subprocess import CalledProcessError, check_call
from mapdamage.version import __version__
import mapdamage
import logging
import time


def construct_path(name,folder="Rscripts"):
    """Construct a path to the mapdamage package data given the name"""
    return(os.path.join(mapdamage.__path__[0], folder, name))

def plot(opt):
  """
  Calls the R script to make the plots, takes in the arguments 
  from parameter processing 
  """
  # comp<-args[1]
  # pdfout<-args[2]
  # around<-as.numeric(args[3])
  # misincorp<-args[4]
  # lg<-as.numeric(args[5])
  # ymax<-as.numeric(args[6])
  # folder<-args[7]
  # title<-args[8]
  # version<-args[9]

  fmut = opt.folder+"/"+"misincorporation.txt"
  fcomp = opt.folder+"/"+"dnacomp.txt"
  title = opt.folder+"/"+"Fragmisincorporation_plot.pdf"

  script = construct_path("mapDamage.R") 
  call = ["Rscript", script, fcomp, title, opt.refplot, fmut, opt.readplot, \
      opt.ymax, opt.folder, opt.title, __version__]
  code = subprocess.call(map(str, call))

  if code == 0:
    if not opt.quiet:
      print("pdf %s generated" % title)
    return 0
  else:
    print("Error: plotting with R failed")
    return 1


        
def opt_plots(opt):       
  """ optional plots for length distribution
  and cumulative C>T mutations, per strand """
  
  fmut = opt.folder+"/"+"misincorporation.txt"
  flength = opt.folder+"/"+"lgdistribution.txt"
  output = opt.folder+"/"+"Length_plot.pdf"

  script = construct_path("lengths.R") 
  call = ["Rscript", script, flength, output, fmut, opt.length, \
      opt.title, __version__, bool_to_int(opt.quiet)]
  code = subprocess.call(map(str, call))
  if code == 0:
    return 0
  else:
    print("Error: plotting with R failed")
    return 1


def check_R_one_lib(name):
    """Checks if a necessary R library is here."""
    with open(os.devnull, "w") as devnull:
        rpa = construct_path("stats/checkLibraries.R")
        return subprocess.call(["Rscript", rpa, "--args", name],
                               stdout = devnull,
                               stderr = devnull)


def check_R_lib():
    """
    Checks if the necessary R libraries are here, signal 
    otherwise
    """
    libs = ["inline", "ggplot2", "gam", "Rcpp", "RcppGSL"]
    missing_lib = []
    for lib in libs:
        # Check the libraries
        if check_R_one_lib(lib):
            #found a missing library
            missing_lib.append(lib)
    if len(missing_lib) > 1:
        # Grammar Nazi has arrived
        last_ele = missing_lib.pop()
        print ("Missing the following R libraries '" + "', '".join(missing_lib) \
            + "' and '" + last_ele + "'")
        return 1
    elif len(missing_lib) == 1:
        print ("Missing the following R library "+missing_lib[0])
        return 1
    else :
        # No missing libraries
        return 0


def run_stats(opt):
    """
    Runs the Bayesian estimation program, using the options opt
    """
    arg = ["Rscript",                            \
         construct_path("stats/runGeneral.R"), \
         "--args",                               \
         opt.rand,                               \
         opt.burn,                               \
         opt.adjust,                             \
         opt.iter,                               \
         bool_to_int(opt.forward),               \
         bool_to_int(opt.reverse),               \
         bool_to_int(not opt.var_disp),              \
         bool_to_int(not opt.diff_hangs),            \
         0,                                      \
         bool_to_int(opt.fix_nicks),             \
         bool_to_int(not opt.single_stranded),       \
         opt.seq_length,                         \
         opt.folder+"/",                         \
         construct_path("stats/"),             \
         opt.folder+"/Stats_out",                \
         int(opt.verbose),                       \
         int(opt.quiet),                         \
         int(opt.jukes_cantor),                       \
         opt.folder+"/acgt_ratio.csv",            \
         int(opt.use_raw_nick_freq),                   \
         int(opt.theme_bw)                   \
         ]
    arg = [str(i) for i in arg]

    logger = logging.getLogger(__name__)
    logger.info("Performing Bayesian estimates")
    logger.debug("Call: %s" % (" ".join(arg),))
    start_time = time.time()

    try:
        check_call(arg)
    except CalledProcessError as e:
        logger.error("The Bayesian statistics program failed to finish")
        raise e

    logger.debug("Bayesian estimates completed in %f seconds" % (time.time() - start_time,))
    return 0


def bool_to_int(boolean):
    """ return 0 for False and 1 for True """
    result = 1 if boolean else 0
    return result



#!/usr/bin/env python

import itertools
import mapdamage

# from Martin Kircher, description of CIGAR operations
#O BAM Description
#M 0 alignment match (can be a sequence match or mismatch)
#I 1 insertion to the reference
#D 2 deletion from the reference
#N 3 skipped region from the reference
#S 4 soft clipping (clipped sequences present in SEQ)
#H 5 hard clipping (clipped sequences NOT present in SEQ)
#P 6 padding (silent deletion from padded reference)
#= 7 sequence match
#X 8 sequence mismatch


def get_coordinates(read):  
  """ return external coordinates of aligned read bases """
  fivep = read.aend if read.is_reverse else read.pos
  threep = read.pos if read.is_reverse else read.aend
  
  return fivep, threep


def get_around(coord, chrom, reflengths, length, ref):
  """ return reference sequences before and after the read
  check for extremities and return what is available """
  coord_min = min(coord)
  coord_max = max(coord)

  pos_before = max(0, coord_min - length)
  pos_after  = min(reflengths[chrom], coord_max + length)

  # Uppercased, to be sure that we don't compare A,C,G,T and a,c,g,t
  before = ref.fetch(chrom, pos_before, coord_min).upper()
  after = ref.fetch(chrom, coord_max, pos_after).upper()

  return before, after


def align(cigarlist, seq, ref):
  """ insert gaps according to the cigar string 
  deletion: gaps to be inserted into read sequences, 
  insertions: gaps to be inserted into reference sequence """  
  lref = list(ref)
  for nbr, idx in parse_cigar(cigarlist, 1):
    lref[idx:idx] = ["-"] * nbr

  lread = list(seq)
  for nbr, idx in parse_cigar(cigarlist, 2):
    lread[idx:idx] = ["-"] * nbr

  return "".join(lread), "".join(lref)


def align_with_qual(cigarlist, seq, qual, threshold, ref):
  """ insert gaps according to the cigar string 
  deletion: gaps to be inserted into read sequences and qualities, 
  insertions: gaps to be inserted into reference sequence """  
  lref = list(ref)
  for nbr, idx in parse_cigar(cigarlist, 1):
    lref[idx:idx] = ["-"] * nbr

  lread = list(seq)
  lqual = list(qual)
  for nbr, idx in parse_cigar(cigarlist, 2):
    lread[idx:idx] = ["-"] * nbr
    lqual[idx:idx] = ["-"] * nbr

  for idx, score in enumerate(lqual):
    # mask read and reference nucleotides (so not gaps) if below desired threshold
    if (ord(score)-33) < threshold and lread[idx] != "-":
      lread[idx] = "N"
      lref[idx]  = "N"

  return "".join(lread), "".join(lqual), "".join(lref)


def get_mis(read, seq, refseq, ref, length, tab, end):
  """ count mismatches using aligned reference and read,
  must be redone since N in reference were randomly replaced by any bases """
  std = '-' if read.is_reverse else '+'
  subtable = tab[ref][end][std]

  for (i, nt_seq, nt_ref) in itertools.izip(xrange(length), seq, refseq):
    if (nt_seq in "ACGT-") and (nt_ref in "ACGT-"):
      if nt_ref != "-":
        # record base composition in the reference, only A, C, G, T
        subtable[nt_ref][i] += 1

      # Most ref/seq pairs will be identical
      if (nt_ref != nt_seq):
        mut = "%s>%s" % (nt_ref, nt_seq)
        subtable[mut][i] += 1


def parse_cigar(cigarlist, ope):
  """ for a specific operation (mismach, match, insertion, deletion... see above)
  return occurences and index in the alignment """
  tlength = 0
  coordinate = []
  # count matches, indels and mismatches
  oplist = (0, 1, 2, 7, 8)
  for operation, length in cigarlist:
    if operation == ope:
        coordinate.append([length, tlength])
    if operation in oplist: 
        tlength += length
  return coordinate


def record_soft_clipping(read, tab, length):
  """ record soft clipped bases at extremities """
  def update_table(end, std, bases):
    for i in range(0, min(bases, length)):
      tab[end][std]['S'][i] += 1

  strand = '-' if read.is_reverse else '+'
  for (nbases, idx) in mapdamage.align.parse_cigar(read.cigar, 4):
    if idx == 0:
      # Soft-clipping at the left side of the alignment
      end = '3p' if read.is_reverse else '5p'
    else:
      # Soft-clipping at the right side of the alignment
      end = '5p' if read.is_reverse else '3p'

    update_table(end, strand, nbases)








import csv
import sys
import os
import mapdamage
import pysam
import itertools
import math
import logging
import time


def phred_pval_to_char(pval):
    """Transforming error rate to ASCII character using the Phred scale"""
    return chr(int(round(-10*math.log10(abs(pval)))+33))

def phred_char_to_pval(ch):
    """Transforming ASCII character in the Phred scale to the error rate"""
    return 10**(-(float(ord(ch))-float(33))/10)

def get_corr_prob(folder, rescale_length_5p, rescale_length_3p):
    """
    Reads the damage probability correction file, returns a 
    dictionary with this structure 
    position (one based)  -  CT  -  probability
                          -  GA  -  probability
    """
    full_path = os.path.join(folder,"Stats_out_MCMC_correct_prob.csv")
    if not os.path.isfile(full_path):
        sys.exit("Missing file, the file \n\tStats_out_MCMC_correct_prob.csv\nshould be in the folder\n\t"+folder+"\nDid you run the MCMC estimates of the parameters?")
    try:
        fi_handle = csv.DictReader(open(full_path))
        corr_prob = {}
        for line in fi_handle:
            if (corr_prob.has_key(line["Position"])):
                sys.exit('This file has multiple position definitions %s, line %d: %s' % \
                    (folder, fi_handle.line_num, corr_prob[line["Position"]]))
            else:
                corr_prob[int(line["Position"])] = {'C.T':float(line["C.T"]), 'G.A':float(line["G.A"])}

        # Exclude probabilities for positions outside of user-specified region
        for key in corr_prob.keys():
            if key < -rescale_length_3p or key > rescale_length_5p:
                corr_prob.pop(key)

        return corr_prob
    except csv.Error as e:
        sys.exit('File %s, line %d: %s' % (os.path.join(folder,"Stats_out_MCMC_correct_prob.csv"), \
            fi_handle.line_num, e))


def corr_this_base(corr_prob, nt_seq, nt_ref, pos, length,direction="both"):
    """
    The position specific damaging correction, using the input 
    corr_prob dictionary holding the damage correcting values 
    nt_seq nucleotide in the sequence 
    nt_ref nucleotide in the reference
    pos relative position from the 5' end 
    length length of the sequence
    direction which end to consider the rescaling
    returns the correction probability for this particular set
    """
    if (pos == 0):
        # not using 0 based indexing
        raise SystemError
    if ( nt_seq == "T" and nt_ref == "C" ):
        # an C to T transition
        subs = "C.T"
    elif( nt_seq == "A" and nt_ref == "G" ):
        # an G to A transition
        subs = "G.A"
    else:
        # other transitions/transversions are not affected by damage
        return 0
    
    back_pos = pos-length-1 
    # position from 3' end

    if corr_prob.has_key(pos):
        p5_corr = corr_prob[pos][subs]
        # correction from 5' end
    else:
        p5_corr = 0

    if corr_prob.has_key(back_pos):
        p3_corr = corr_prob[back_pos][subs]
        # correction from 3' end
    else:
        p3_corr = 0

    if direction == "forward":
        return p5_corr
    elif direction == "backward":
        return p3_corr 
    elif direction == "both":
        if pos < abs(back_pos) :
            # then we use the forward correction
            return p5_corr
        else :
            # else the backward correction
            return p3_corr
    else:
        # this should not happen
        raise SystemExit("Abnormal direction in the rescaling procedure")

def initialize_subs():
    """Initialize a substitution table, to track the expected substitution counts"""
    per_qual = dict(zip(range(0,130),[0]*130))
    subs = {"CT-before":per_qual.copy(),\
            "TC-before":per_qual.copy(),\
            "GA-before":per_qual.copy(),\
            "AG-before":per_qual.copy(),\
            "CT-after":per_qual.copy(),\
            "TC-after":per_qual.copy(),\
            "GA-after":per_qual.copy(),\
            "AG-after":per_qual.copy(),\
            "A":0,\
            "C":0,\
            "G":0,\
            "T":0,\
            "CT-pvals":0.0,\
            "CT-pvals_before":0.0,\
            "TC-pvals":0.0,\
            "GA-pvals":0.0,\
            "GA-pvals_before":0.0,\
            "AG-pvals":0.0,\
            }
    return subs



def record_subs(subs,nt_seq,nt_ref,nt_qual,nt_newqual,prob_corr):
    """ record the expected substitution change, prob_corr is the excact version for nt_qual"""
    if ( nt_seq == "T" and nt_ref == "C"):
        sub_type = "CT"
        subs["CT-pvals"] += prob_corr
        subs["CT-pvals_before"] += 1-phred_char_to_pval(nt_qual)
    elif ( nt_seq == "A" and nt_ref == "G"):
        sub_type = "GA"
        subs["GA-pvals"] += prob_corr
        subs["GA-pvals_before"] += 1-phred_char_to_pval(nt_qual)
    elif ( nt_seq == "C" and nt_ref == "T"):
        sub_type = "TC"
        subs["TC-pvals"] += 1-phred_char_to_pval(nt_qual)
        if (nt_qual != nt_newqual):
            raise SystemError("Internal error: rescaling qualities for the wrong transitions")
    elif ( nt_seq == "G" and nt_ref == "A"):
        sub_type = "AG"
        subs["AG-pvals"] += 1-phred_char_to_pval(nt_qual)
        if (nt_qual != nt_newqual):
            raise SystemError("Internal error: rescaling qualities for the wrong transitions")
    else:
        sub_type = "NN"
    if (sub_type != "NN"):
        # record only transitions 
        subs[sub_type+"-before"][int(ord(nt_qual))-33] += 1
        subs[sub_type+"-after"][int(ord(nt_newqual))-33] += 1
    if (nt_ref in ["A","C","G","T"]):
        subs[nt_ref] += 1

def qual_summary_subs(subs):
    """Calculates summary statistics for the substition table subs"""
    for i in ["CT-before","TC-before","GA-before","AG-before","CT-after","TC-after","GA-after","AG-after"]:
        for lv in [0,10,20,30,40]:
            for qv in subs[i]:
                if qv >= lv :
                    key = i+"-Q"+str(lv)
                    if subs.has_key(key):
                        subs[key] += subs[i][qv]
                    else:
                        subs[key] = subs[i][qv]

def print_subs(subs):
    """Print the substition table"""
    print("\tThe expected substition frequencies before and after scaling using the scaled qualities as probalities:")
    if subs["C"]!=0:
        # the special case of no substitutions
        print("\tCT\t"+str(subs["CT-pvals_before"]/subs["C"])+"\t\t"+str(subs["CT-pvals"]/subs["C"]))
    else: 
        print("\tCT\tNA\t\tNA")
    if subs["T"]!=0:
        print("\tTC\t"+str(subs["TC-pvals"]/subs["T"])+"\t\t"+str(subs["TC-pvals"]/subs["T"]))
    else:
        print("\tTC\tNA\t\tNA")
    if subs["G"]!=0:
        print("\tGA\t"+str(subs["GA-pvals_before"]/subs["G"])+"\t\t"+str(subs["GA-pvals"]/subs["G"]))
    else:
        print("\tGA\tNA\t\tNA")
    if subs["A"]!=0:
        print("\tAG\t"+str(subs["AG-pvals"]/subs["A"])+"\t\t"+str(subs["AG-pvals"]/subs["A"]))
    else:
        print("\tAG\tNA\t\tNA")
    print("\tQuality metrics before and after scaling")
    print("\tCT-Q0 \t"+str(subs["CT-before-Q0"])+"\t\t"+str(subs["CT-after-Q0"]))
    print("\tCT-Q10 \t"+str(subs["CT-before-Q10"])+"\t\t"+str(subs["CT-after-Q10"]))
    print("\tCT-Q20 \t"+str(subs["CT-before-Q20"])+"\t\t"+str(subs["CT-after-Q20"]))
    print("\tCT-Q30 \t"+str(subs["CT-before-Q30"])+"\t\t"+str(subs["CT-after-Q30"]))
    print("\tCT-Q40 \t"+str(subs["CT-before-Q40"])+"\t\t"+str(subs["CT-after-Q40"]))
    print("\tGA-Q0 \t"+str(subs["GA-before-Q0"])+"\t\t"+str(subs["GA-after-Q0"]))
    print("\tGA-Q10 \t"+str(subs["GA-before-Q10"])+"\t\t"+str(subs["GA-after-Q10"]))
    print("\tGA-Q20 \t"+str(subs["GA-before-Q20"])+"\t\t"+str(subs["GA-after-Q20"]))
    print("\tGA-Q30 \t"+str(subs["GA-before-Q30"])+"\t\t"+str(subs["GA-after-Q30"]))
    print("\tGA-Q40 \t"+str(subs["GA-before-Q40"])+"\t\t"+str(subs["GA-after-Q40"]))
    

def rescale_qual_read(bam, read, ref, corr_prob,subs, debug = False,direction="both"):
    """
    bam              a pysam bam object
    read             a pysam read object
    ref              a pysam fasta ref file
    reflengths       a dictionary holding the length of the references 
    subs             a dictionary holding the corrected number of substition before and after scaling 
    corr_prob dictionary from get_corr_prob
    returns a read with rescaled quality score
    
    Iterates through the read and reference, rescales the quality 
    according to corr_prob
    """
    if not debug:
        # no need to log when unit testing
        logger = logging.getLogger(__name__)
    raw_seq = read.query
    # external coordinates 5' and 3' , 0-based offset
    coordinate = mapdamage.align.get_coordinates(read)
    # fetch reference name, chromosome or contig names
    chrom = bam.getrname(read.tid)
    refseq = ref.fetch(chrom, min(coordinate), max(coordinate)).upper()
    # add gaps to qualities and mask read and reference nucleotides if below desired threshold
    (seq, qual, refseq) = mapdamage.align.align_with_qual(read.cigar, \
        raw_seq, read.qqual, -100, refseq)
    length_read = len(raw_seq)
    length_align = len(seq)
    # reverse complement read and reference when mapped reverse strand
    if read.is_reverse:
        refseq = mapdamage.seq.revcomp(refseq)
        seq = mapdamage.seq.revcomp(seq)
        qual = qual[::-1]
    new_qual = [-100]*length_read
    pos_on_read = 0
    number_of_rescaled_bases = 0.0
    for (i, nt_seq, nt_ref, nt_qual) in itertools.izip(xrange(length_align), seq, refseq, qual):
        # rescale the quality according to the triplet position, 
        # pair of the reference and the sequence
        if ((nt_seq == "T" and nt_ref =="C") or (nt_seq == "A" and nt_ref =="G")):
            # need to rescale this subs.
            pdam = 1 - corr_this_base(corr_prob, nt_seq, nt_ref, pos_on_read + 1, length_read,direction=direction)
            pseq = 1 - phred_char_to_pval(nt_qual)
            newp = pdam*pseq # this could be numerically unstable
            newq = phred_pval_to_char(1-newp)
            number_of_rescaled_bases += 1-pdam
        else:
            # don't rescale, other bases
            newp = 1 - phred_char_to_pval(nt_qual)
            newq = nt_qual 
        if pos_on_read < length_read:
            new_qual[pos_on_read] = newq 
            record_subs(subs,nt_seq,nt_ref,nt_qual,new_qual[pos_on_read],newp)
            if nt_seq != "-":
                pos_on_read += 1
            # done with the aligned portion of the read 
        else:
            if not debug:
                logger.warning("Warning: The aligment of the read is longer than the actual read %s",(read.qname))
            break
    new_qual = "".join(new_qual)

    if read.is_reverse:
        new_qual = new_qual[::-1]
    if (read.cigar[0][0] == 4):
        # check for soft clipping at forward end 
        new_qual = read.qual[0:read.cigar[0][1]] + new_qual
    if (read.cigar[-1][0] == 4):
        # the same backwards
        new_qual = new_qual + read.qual[-read.cigar[-1][1]:]

    read.qual = new_qual
    # truncate this to 5 digits
    number_of_rescaled_bases = float("%.5f" % number_of_rescaled_bases)

    if read.has_tag("MR"):
        raise SystemExit("Read: %s already has a MR tag, can't rescale" % read)

    read.set_tag("MR", number_of_rescaled_bases, 'f')

    return read


def rescale_qual(ref, options,debug=False):
    """    
    ref                a pysam fasta ref file
    bam_filename       name of a BAM/SAM file to read
    fi                 file containing the csv with correction probabilities
    reflengths         dictionary with the reference lengths
    options            options from the command line parsing

    Iterates through BAM file, makes a new BAM file with rescaled qualities.
    """
    if not debug:
        # no need to log when unit testing
        logger = logging.getLogger(__name__)
        logger.info("Rescaling BAM: '%s' -> '%s'" % (options.filename, options.rescale_out))
    start_time = time.time()

    # open SAM/BAM files
    bam = pysam.Samfile(options.filename)
    if debug:
        write_mode = "wh"
    else:
        write_mode = "wb"
    bam_out = pysam.Samfile(options.rescale_out,write_mode, template = bam)
    corr_prob = get_corr_prob(options.folder,
                              rescale_length_5p=options.rescale_length_5p,
                              rescale_length_3p=options.rescale_length_3p)
    subs = initialize_subs()
    first_pair = True
    number_of_non_proper_pairs = 0

    for hit in bam:
        if hit.is_unmapped:
            pass
        elif not hit.qual and not debug:
            logger.warning("Cannot rescale base PHRED scores for read '%s'; no scores assigned." % hit.qname)
        elif hit.is_paired : 
            if first_pair and not debug:
                # assuming the ends are non-overlapping 
                logger.warning("Warning! Assuming the pairs are non-overlapping, facing inwards and correctly paired.")
                first_pair=False
            #5p --------------> 3p
            #3p <-------------- 5p
            # pair 1 (inwards)
            #5p ----> 
            #             <---- 5p
            #     A         B
            # pair 2 (outwards), this happens if the reference is RC this is not supported 
            #             ----> 3p
            #3p <----         
            #     A         B
            # Correct outwards pairs from the 3p and inwards pairs with the 5p end
            if ((not hit.is_reverse) and hit.mate_is_reverse and (hit.pnext>hit.pos) and hit.tid==hit.mrnm):
                # the inwards case mate A
                hit = rescale_qual_read(bam, hit, ref, corr_prob,subs,direction="forward",debug=debug)
            elif (hit.is_reverse and (not hit.mate_is_reverse) and (hit.pnext<hit.pos) and hit.tid==hit.mrnm):
                # the inwards case mate B
                hit = rescale_qual_read(bam, hit, ref, corr_prob,subs,direction="forward",debug=debug)
            else:
                number_of_non_proper_pairs += 1
                # cannot do much with conflicting pairing information
        else:
            hit = rescale_qual_read(bam, hit, ref, corr_prob,subs,debug=debug)

        bam_out.write(hit)
    if  number_of_non_proper_pairs!=0 and not debug:
        logger.warning("Number of non-rescaled reads due to improper pairing:  %d" % number_of_non_proper_pairs)
    if (subs["TC-before"] != subs["TC-after"] or subs["AG-before"] != subs["AG-after"]):
        sys.exit("Qualities for T.C and A.G transitions should not change in the rescaling, please contact the authors.")
    qual_summary_subs(subs)
    bam.close()
    bam_out.close()
    if not options.quiet:
        print_subs(subs)
    if not debug:
        logger.debug("Rescaling completed in %f seconds" % (time.time() - start_time,))
